from yo_fluq import *
from unittest import TestCase



class LittleMethodsTests(TestCase):

    def test_min(self):
        self.assertEqual(1,Query.args(1,2,3).min())

    def test_min_on_empty_with_default(self):
        self.assertIsNone(Query.args().min(default=None))

    def test_min_on_empty(self):
        self.assertRaises(ValueError,lambda: Query.args().min())

    def test_min_with_selector(self):
        self.assertEqual(-3,Query.args(1,2,3).min(lambda z: -z))


    def test_max(self):
        self.assertEqual(3,Query.args(1,2,3).max())

    def test_max_on_empty_with_default(self):
        self.assertIsNone(Query.args().max(default=None))

    def test_max_on_empty(self):
        self.assertRaises(ValueError,lambda: Query.args().max())

    def test_max_with_selector(self):
        self.assertEqual(-1,Query.args(1,2,3).max(lambda z: -z))


    def test_argmax_1(self):
        self.assertEqual(-1, Query.args(-1, 0, 1).argmax(lambda z: abs(z)))

    def test_argmax_2(self):
        self.assertRaises(ValueError,lambda: Query.args().argmax(lambda z: abs(z)))

    def test_argmax_or_default_1(self):
        self.assertEqual(-1, Query.args(-1, 0, 1).argmax(lambda z: abs(z), default=10))

    def test_argmax_or_default_2(self):
        self.assertEqual(-10, Query.args().argmax(lambda z: abs(z), default = -10))

    def test_argmin_1(self):
        self.assertEqual(0, Query.args(-1, 0, 1).argmin(lambda z: abs(z)))

    def test_argmin_2(self):
        self.assertRaises(ValueError,lambda: Query.args().argmin(lambda z: abs(z)))


    def test_argmin_or_default_1(self):
        self.assertEqual(0, Query.args(-1, 0, 1).argmin(lambda z: abs(z),default=-10))

    def test_argmin_or_default_2(self):
        self.assertEqual(-10, Query.args().argmin(lambda z: abs(z), -10))




from yo_fluq import *
from unittest import TestCase


class AggregationMethodsTests(TestCase):


    def test_count(self):
        self.assertEqual(3,Query.args(1,2,3).count())

    def test_count_float(self):
        t = Query.args(1,2,3).aggregate_with(agg.Count().astype(float))
        self.assertIsInstance(t,float)
        self.assertEqual(3.0,t)

    def test_sum(self):
        self.assertEqual(6,Query.args(1,2,3).sum())

    def test_sum_on_empty(self):
        self.assertRaises(ValueError, lambda: Query.args().sum())

    def test_sum_on_empty_default(self):
        self.assertIsNone(Query.args().aggregate_with(agg.Sum().default_if_empty()))

    def test_mean(self):
        self.assertEqual(2.0,Query.args(1,2,3).mean())

    def test_mean_on_empty(self):
        self.assertIsNone(Query.args().mean())

    def test_std(self):
        arr = [0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777]
        std_expected = 0.18447489445957163
        self.assertAlmostEqual(std_expected,Query.en(arr).aggregate_with(agg.Std()),5)

    def test_std_single(self):
        self.assertIsNone(Query.args(1).aggregate_with(agg.Std()))





import time
from yo_fluq import *
from unittest import TestCase
import os
from typing import *

def _time_consuming_selector(q): # pragma: no cover
    time.sleep(0.1)
    return q


class to_text_file(Callable[[Queryable],None]):
    time.sleep(0.1)
    def __init__(self, filename: str, autoflush: bool = False, separator: str ='\n', **kwargs): # pragma: no cover
        self.filename = filename
        self.autoflush = autoflush
        self.separator = separator
        self.kwargs = kwargs

    def __call__(self, query: Queryable):  # pragma: no cover
        with open(self.filename, 'w', **self.kwargs) as file:
            for item in query:
                file.write(str(item) + self.separator)
                if self.autoflush:
                    file.flush()

def _text_file(filename, **kwargs): # pragma: no cover
    with open(filename, 'r', **kwargs) as file:
        for line in file:
            if line.endswith('\n'):
                line = line[:-1]
            yield line



class ParallelTests(TestCase):
    def test_fork(self):
        ctx = ForkContext(dict(factor=2, buffer = []))
        full_result = (
                Query
                .en(range(15))
                .fork(ctx, lambda query, context: Query
                      .en(query)
                      .where(lambda z: z%context['factor']==0)
                      .foreach_and_continue(lambda z: context['buffer'].append(z))
                      .select(lambda z: -z)
                      .to_list())
                .to_list()
        )
        self.assertListEqual(list(range(15)),full_result)
        self.assertListEqual([], ctx.sent['buffer'])
        self.assertListEqual([z for z in range(15) if z%2==0], ctx.received['buffer'])
        self.assertListEqual([-z for z in range(15) if z%2==0], ctx.result)


    def test_fire_and_forget(self):
        path = os.path.join(os.path.dirname(__file__),'temporary.parallel.test.txt')
        if os.path.isfile(path):  # pragma: no cover
            os.remove(path)
        full_result = (Query
                       .en(range(15))
                       .fire_and_forget(lambda q: Query.en(q).where(lambda z: z%2==0).feed(to_text_file(path)))
                       .to_list()
                       )
        self.assertListEqual(list(range(15)),full_result)
        self.assertListEqual([z for z in range(15) if z%2==0], Queryable(_text_file(path)).select(int).to_list())

    def test_parallel_select_0(self):
        start_time = time.time()
        result = Query.en(range(10)).parallel_select(_time_consuming_selector,4,1).to_list()
        duration = time.time()-start_time
        self.assertLess(duration,0.5)




from .queryable import Queryable
from .._push_queries import PushQuery
from typing import *
from collections import Sized
from .._common import *
from .helpers import loop_maker

class QueryClass:
    def en(self, en: Iterable) -> Queryable:
        length = None
        if isinstance(en,Sized):
            length = len(en)
        return FlupFactory.QueryableFactory(en,length)

    def args(self, *args)  -> Queryable:
        return FlupFactory.QueryableFactory(args,len(args))


    def dict(self, dictionary: Dict)  -> Queryable:
        return FlupFactory.QueryableFactory(dictionary.items(),len(dictionary)).select(lambda z: KeyValuePair(z[0],z[1]))

    def push(self) -> PushQuery:
        return PushQuery()

    def loop(self, begin: Any, delta: Any, end: Any = None, endtype=LoopEndType.NotEqual):
        lp = loop_maker(begin,delta,end,endtype)
        return Queryable(lp.make())




from yo_fluq import *
from unittest import TestCase


class ToCollectionMethodsTests(TestCase):
    def test_to_dictionary_normal(self):
        d = Query.args(1,2,3).to_dictionary(lambda z: z, lambda z: str(z))
        self.assertDictEqual({1:'1',2:'2',3:'3'},d)

    def test_to_dictionary_duplicating(self):
        self.assertRaises(ValueError,lambda: Query.args(1,1,2).to_dictionary(lambda z: z, lambda z: str(z)))

    def test_to_dictionary_key_value_pair(self):
        d = Query.args(1,2,3).select(lambda z: KeyValuePair(z,str(z))).to_dictionary()
        self.assertDictEqual({1: '1', 2: '2', 3: '3'}, d)

    def test_to_dictionary_key_value_pair_override(self):
        d = Query.dict(dict(a=1,b=2)).to_dictionary(lambda z: z.key.upper(), lambda z: z.value+1)
        self.assertDictEqual(dict(A=2,B=3),d)

    def test_to_dictionary_mixture(self):
        self.assertRaises(ValueError,lambda: Query.args(KeyValuePair(1,2),1).to_dictionary())

    def test_to_dictionary_without_key_selector(self):
        self.assertRaises(ValueError, lambda: Query.args(1).to_dictionary())

    def test_to_dictionary_without_value_selector(self):
        self.assertRaises(ValueError, lambda: Query.args(1).to_dictionary(lambda z: z))

    def test_to_dictionary_grby(self):
        result = Query.args(1,2,3,4).group_by(lambda z: z%2).to_dictionary()
        self.assertDictEqual({0:[2,4],1:[1,3]},result)

    def test_to_tuple(self):
        t = Query.args(1, 2, 3).to_tuple()
        self.assertIsInstance(t, tuple)
        self.assertListEqual([1, 2, 3], list(t))

    def test_to_set_1(self):
        st = Query.args(1, 2, 3).to_set()
        self.assertSetEqual({1, 2, 3}, st)

    def test_to_set_2(self):
        self.assertRaises(ValueError, lambda: Query.args(1, 2, 2).to_set())

